# Adapted in part from Olivier Hagolle
# https://github.com/olivierhagolle/LANDSAT-Download
import os
import re
import requests
from future.standard_library import install_aliases
import tarfile
from pandas import read_csv

install_aliases()
from urllib.parse import urlencode
from urllib.request import urlopen, install_opener, build_opener, HTTPCookieProcessor, Request


class StationNotFoundError(Exception):
    pass


class InvalidSatelliteError(Exception):
    pass


class BadRequestsResponse(Exception):
    pass


class InvalidCredentialsResponse(Exception):
    pass


def download_image(url, output_dir, image, creds):
    cookies = HTTPCookieProcessor()
    opener = build_opener(cookies)
    install_opener(opener)

    data = urlopen("https://ers.cr.usgs.gov").read().decode('utf-8')
    m = re.search(r'<input .*?name="csrf_token".*?value="(.*?)"', data)
    if m:
        token = m.group(1)
    else:
        print("Error : CSRF_Token not found")

    params = urlencode(dict(username=creds['account'], password=creds['passwd'], csrf_token=token))
    params = params.encode('ascii')
    request = Request("https://ers.cr.usgs.gov/login", params, headers={})
    f = urlopen(request)

    data = f.read().decode('utf-8')
    f.close()
    if data.find('Invalid username/password') > 0:
        print("Authentification failed")
        raise InvalidCredentialsResponse('Check your credentials, '
                                         'they were rejected by USGS')

    req = urlopen(url)

    uri = req.url
    response = requests.get(uri)

    if response.status_code == 200:
        with open(os.path.join(output_dir, image), 'wb') as f:
            for chunk in response.iter_content(chunk_size=1024 * 1024 * 8):
                f.write(chunk)

    elif response.status_code > 399:
        print('Code {}'.format(response.status_code))
        raise BadRequestsResponse(Exception)


def unzip_image(tgzfile, outputdir):
    target_tgz = os.path.join(outputdir, tgzfile)
    if os.path.exists(target_tgz):
        print('found tgs: {} \nunzipping...'.format(target_tgz))
        tfile = tarfile.open(target_tgz, 'r:gz')
        tfile.extractall(outputdir)
        print('unzipped\ndeleting tgz: {}'.format(target_tgz))
        os.remove(target_tgz)
    else:
        raise NotImplementedError('Did not find download output directory to unzip...')
    return None


def get_credentials(usgs_path):
    with open(usgs_path) as f:
        (account, passwd) = f.readline().split(' ')
        if passwd.endswith('\n'):
            passwd = passwd[:-1]
        usgs = {'account': account, 'passwd': passwd}
        return usgs


def get_station_list_identifier(product):
    if product.startswith('LC8'):
        identifier = '12864'
        stations = ['LGN']
    elif product.startswith('LE7'):
        identifier = '12267'
        stations = ['EDC', 'SGS', 'AGS', 'ASN', 'SG1', 'CUB', 'COA']
    elif product.startswith('LT5'):
        identifier = '12266'
        stations = ['GLC', 'ASA', 'KIR', 'MOR', 'KHC', 'PAC',
                    'KIS', 'CHM', 'LGS', 'MGR', 'COA', 'MPS', 'CUB']
    else:
        raise NotImplementedError('Must provide valid product string...')

    return identifier, stations


def get_candidate_scenes_list(path, row, sat_name, start_date, end_date, max_cloud_cover=70):
    """
    
    :param path: path (int) 
    :param row: row (int)
    :param sat_name: 'LT5', 'LE7', or 'LC8'
    :param start_date: datetime object start image search
    :param end_date: datetime object finish image search
    :param max_cloud_cover: percent cloud cover according to USGS image metadata, float
    :return: reference overpass = str('YYYYDOY'), station str('XXX') len=3
    """
    sensor_map = {'LT5': 'LANDSAT_TM_C1', 'LE7': 'LANDSAT_ETM_C1', 'LC8': 'LANDSAT_8_C1'}
    sensor = sensor_map[sat_name]
    start = start_date.strftime('%Y-%m-%d')
    end = end_date.strftime('%Y-%m-%d')
    meta_url = 'https://earthexplorer.usgs.gov/EE/InventoryStream/'
    query = 'pathrow?start_path={a}&end_path={a}&start_row={b}&end_row={b}&sensor={c}&' \
            'start_date={d}&end_date={e}&format=CSV'.format(a=path,
                                                            b=row,
                                                            c=sensor,
                                                            d=start,
                                                            e=end)
    url = '{}{}'.format(meta_url, query)
    csv = read_csv(url, header=0)
    csv = csv[csv['cloudCoverFull'] <= max_cloud_cover]
    scene_list = csv['sceneID'].tolist()

    return scene_list


def down_usgs_by_list(scene_list, output_dir, usgs_creds_txt, zipped=False):
    usgs_creds = get_credentials(usgs_creds_txt)

    for product in scene_list:
        print(product)
        identifier, stations = get_station_list_identifier(product)
        base_url = 'https://earthexplorer.usgs.gov/download/'
        tail_string = '{}/{}/STANDARD/EE'.format(identifier, product)
        url = '{}{}'.format(base_url, tail_string)
        tgz_file = '{}.tgz'.format(product)
        scene_dir = os.path.join(output_dir, product)
        if not os.path.isdir(scene_dir):
            os.mkdir(scene_dir)
        if len(os.listdir(scene_dir)) < 1:
            download_image(url, scene_dir, tgz_file, usgs_creds)
            print('image: {}'.format(os.path.join(scene_dir, tgz_file)))
            if not zipped:
                unzip_image(tgz_file, scene_dir)
        else:
            print('This image already exists at {}'.format(output_dir))

    return None


if __name__ == '__main__':
    pass

# ===============================================================================



#!/usr/bin/env python

# ===============================================================================
# Copyright 2017 dgketchum
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ===============================================================================
import os
import argparse
import sys
import yaml
from datetime import datetime

try:
    from core.download_composer import download_landsat
except ImportError:
    sys.path.append(os.path.dirname(os.path.dirname(os.path.abspath(__file__))))
    from core.download_composer import download_landsat


class TooFewInputsError(Exception):
    pass


DEFAULT_CFG = '''
# date format: 'YYYY-MM-DD'
start: '2007-05-01'
end: '2007-05-31'
path: 43
row: 30
latitude:
longitude:
output_path: None
satellite: LT5
usgs_creds: /path/to/usgs_creds.txt
return_list: False
zipped: False
max_cloud_percent: 100
'''


def create_parser():
    parser = argparse.ArgumentParser(prog='landsat', description='Download and unzip landsat data.')

    parser.add_argument('--satellite', help='Satellite name: LT5, LE7, or LC8')
    parser.add_argument('--start', help='Start date in format YYYY-MM-DD')
    parser.add_argument('--end', help='End date in format YYYY-MM-DD')
    parser.add_argument('-lat', '--latitude', help='Latitude, decimal degrees', type=str)
    parser.add_argument('-lon', '--longitude', help='Longitude, decimal degrees', type=str)
    parser.add_argument('-p', '--path', help='The path')
    parser.add_argument('-r', '--row', help='The row')
    parser.add_argument('-o', '--output', help='Output directory', default=os.getcwd())
    parser.add_argument('-cred', '--credentials',
                        help='Path to a text file with USGS credentials with one space between <username password>')

    parser.add_argument('-conf', '--configuration', help='Path to your configuration file. If a directory is provided,'
                                                         'a template cofiguration file will be created there.')

    parser.add_argument('--return-list', help='Just return list of images without downloading', action='store_true',
                        default=False)

    parser.add_argument('--zipped', help='Download .tar.gz file(s), without unzipping',
                        action='store_true', default=False)

    parser.add_argument('--max-cloud-percent', help='Maximum percent of of image obscured by clouds accepted,'
                                                    ' type integer',
                        default=None)

    return parser


def main(args):
    if args:
        print(args)

        fmt = '%Y-%m-%d'

        cfg = {'output_path': args.output,
               'usgs_creds': args.credentials,
               'return_list': args.return_list,
               'zipped': args.zipped}

        if args.configuration:
            if os.path.isdir(args.configuration):
                print('Creating template configuration file at {}.'.format(args.configuration))
                check_config(args.configuration)

            print('\nStarting download with configuration file {}'.format(args.configuration))
            with open(args.configuration, 'r') as rfile:
                ycfg = yaml.load(rfile)
                cfg.update(ycfg)
            cfg['start'] = datetime.strptime(cfg['start'], fmt)
            cfg['end'] = datetime.strptime(cfg['end'], fmt)
            scenes = download_landsat(**cfg)

        if not args.configuration:
            if not args.start:
                raise TooFewInputsError('Must specify path and row, or latitude and longitude, '
                                        'or the path to file with one of these pairs')
            start = datetime.strptime(args.start, fmt)
            end = datetime.strptime(args.end, fmt)
            sat = args.satellite

            if args.latitude:
                print('\nStarting download with latlon...')
                cfg['latitude'] = args.latitude
                cfg['longitude'] = args.longitude
                scenes = download_landsat(start, end, sat, **cfg)

            elif args.path:
                print('\nStarting download with pathrow...')
                cfg['path'] = args.path
                cfg['row'] = args.row
                scenes = download_landsat(start, end, sat, **cfg)

        if args.return_list:
            return scenes


def cli_runner():
    parser = create_parser()
    args = parser.parse_args()
    return main(args)


def check_config(dirname):
    path = os.path.join(dirname, 'downloader_config.yml')
    print('\n*****A default config file {} will be written'.format(path))

    with open(path, 'w') as wfile:
        print('-------------- DEFAULT CONFIG -----------------')
        print(DEFAULT_CFG)
        print('-----------------------------------------------')
        wfile.write(DEFAULT_CFG)

    print('***** Please edit the config file at {} and run the downer again *****\n'.format(
        dirname))

    sys.exit()


if __name__ == '__main__':
    cli_runner()

# ===============================================================================



# ===============================================================================
# Copyright 2017 dgketchum
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ===============================================================================
import re
import requests
from lxml import html


class OverpassNotFoundError(Exception):
    pass


class InvalidDateForSatelliteError(Exception):
    pass


class WebSiteOfflineError(Exception):
    pass


def convert_lat_lon_wrs2pr(lat, lon, conversion_type='convert_ll_to_pr'):
    """ 
    :param lat: Latitude float
    :param lon: Longitude float
    :param conversion_type: 'convert_ll_to_pr' [coordinates to path, row]; 
            'convert_pr_to_ll' [path, row to coordinates]
    :return: lat, lon tuple or path, row tuple
    """
    base = 'https://landsat.usgs.gov/landsat/lat_long_converter/tools_latlong.php'
    # I never figured out how to get unk_number programatically, and it changes,
    # unknown parameter purpose
    unk_number = 1508518830987

    if conversion_type == 'convert_ll_to_pr':

        full_url = '{}?rs={}&rsargs[]={}&rsargs[]={}&rsargs[]=1&rsrnd={}'.format(base, conversion_type,
                                                                                 lat, lon,
                                                                                 unk_number)
        r = requests.get(full_url)
        tree = html.fromstring(r.text)

        # remember to view source html to build xpath
        # i.e. inspect element > network > find GET with relevant PARAMS
        # > go to GET URL > view source HTML
        p_string = tree.xpath('//table/tr[1]/td[2]/text()')
        path = int(re.search(r'\d+', p_string[0]).group())

        r_string = tree.xpath('//table/tr[1]/td[4]/text()')
        row = int(re.search(r'\d+', r_string[0]).group())

        print('path: {}, row: {}'.format(path, row))

        return path, row
    elif conversion_type == 'convert_pr_to_ll':

        full_url = '{}?rs={}&rsargs[]=\n' \
                   '{}&rsargs[]={}&rsargs[]=1&rsrnd={}'.format(base, conversion_type,
                                                               lat, lon, unk_number)
        # this needs to depend on something other than USGS
        r = requests.get(full_url)
        tree = html.fromstring(r.text)

        lat_string = tree.xpath('//table/tr[2]/td[2]/text()')
        lat = float(re.search(r'[+-]?\d+(?:\.\d+)?', lat_string[0]).group())

        lon_string = tree.xpath('//table/tr[2]/td[4]/text()')
        lon = float(re.search(r'[+-]?\d+(?:\.\d+)?', lon_string[0]).group())

        return lat, lon

    else:
        raise NotImplementedError('Must chose either convert_pr_to_ll or convert_ll_to_pr')


if __name__ == '__main__':
    pass

# ==================================================================================



# ===============================================================================
# Copyright 2017 dgketchum
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ===============================================================================


import os
from datetime import datetime

from core import usgs_download
from core.web_tools import convert_lat_lon_wrs2pr


class InvalidPathRowData(Exception):
    pass


def download_landsat(start=None, end=None, satellite=None, latitude=None, longitude=None,
                     path=None, row=None, output_path=None,
                     usgs_creds=None, return_list=False, zipped=False, max_cloud_percent=70):
    if path:
        pass

    elif latitude and longitude:
        path, row = convert_lat_lon_wrs2pr(latitude, longitude)

    else:
        raise InvalidPathRowData('Must give path/row tuple, lat/lon tuple plus row/path \n'
                                 'shapefile, or a path/rows shapefile!')

    scenes_list = usgs_download.get_candidate_scenes_list(path=path, row=row, sat_name=satellite, start_date=start,
                                                          end_date=end, max_cloud_cover=max_cloud_percent)
    if not scenes_list:
        print('No scenes for {} between {} and {}.'.format(satellite,
                                                           datetime.strftime(start,
                                                                             '%Y-doy %j'),
                                                           datetime.strftime(end,
                                                                             '%Y-doy %j')))
        return None

    elif return_list:

        print(scenes_list)

        return scenes_list

    else:

        destination_path = os.path.join(output_path, '{}_{}_{}'.format(
            satellite, path, row))

        if not os.path.exists(destination_path):
            print('making dir: {}'.format(destination_path))
            os.makedirs(destination_path)

        usgs_download.down_usgs_by_list(scenes_list, destination_path,
                                        usgs_creds, zipped)

        return None


if __name__ == 'main':
    pass

# ===============================================================================
